# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import json
import re
import socket
import traceback
import urllib
import urllib2

from django.conf import settings

from identityprovider.timeline_helpers import maybe_contextmanager


class CaptchaError(Exception):
    """All instances of CaptchaError have a traceback member."""

    def __init__(self, traceback):
        self.traceback = traceback


class VerifyCaptchaError(CaptchaError):
    def __init__(self, response):
        super(VerifyCaptchaError, self).__init__(response.traceback)
        self.response = response


class EmptyCaptchaKey(RuntimeError):
    pass


class CaptchaResponse(object):
    """Response returned from _open() in case of error"""

    def __init__(self, code, response, tb=None, reason=None, body=None):
        self.code = code
        self.response = response
        self.traceback = tb
        self.reason = reason
        self.body = body
        self.is_error = tb is not None
        self._data = None

    def data(self):
        if self._data:
            return self._data
        if self.response:
            self._data = self.response.read()
            return self._data
        else:
            return None


class CaptchaBase(object):
    """ABC to capture & abstract interaction with external captcha service."""

    opener = None

    def __init__(self):
        self._setup_opener()

    @classmethod
    def _setup_opener(cls):
        if cls.opener is not None:
            return
        if settings.CAPTCHA_PROXIES:
            proxy_handler = urllib2.ProxyHandler(settings.CAPTCHA_PROXIES)
            opener = urllib2.build_opener(proxy_handler)
        else:
            opener = urllib2.build_opener()
        cls.opener = opener

    @classmethod
    def _open(cls, request):
        default_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(getattr(settings, 'CAPTCHA_TIMEOUT', 10))

        try:
            response = cls.opener.open(request)

        except urllib2.HTTPError as e:
            if e.code not in (408, 500, 502, 503, 504):
                # 408: Request Timeout
                # 500: Internal Server Error
                # 502: Bad Gateway
                # 503: Service Unavailable
                # 504: Gateway Timeout
                raise
            tb = traceback.format_exc()
            return CaptchaResponse(e.code, None, tb, e.reason, e.read())

        except urllib2.URLError as e:
            error_code = e.reason[0]
            if error_code not in (111, 113):
                # tcp errors
                # 111: Connection refused
                # 113: No route to host
                raise
            tb = traceback.format_exc()
            return CaptchaResponse(error_code, None, tb, e.reason[1])

        finally:
            socket.setdefaulttimeout(default_timeout)

        return CaptchaResponse(response.code, response, None)

    def check_whitelist(self, email):
        for pattern in settings.EMAIL_WHITELIST_REGEXP_LIST:
            if re.match(pattern, email):
                return True
        return False


class CaptchaV2(CaptchaBase):
    """Class to capture & abstract interaction with Google reCaptcha v2.0."""

    def __init__(self, timer=None):
        self.opener = None
        self._verified = None
        with maybe_contextmanager(timer, 'captcha-new', None):
            super(CaptchaV2, self).__init__()

    def verify(self, captcha_solution, remote_ip, email, timer=None):
        if self._verified is not None:
            return self._verified

        if self.check_whitelist(email):
            return True

        if not settings.CAPTCHA_PRIVATE_KEY:
            raise EmptyCaptchaKey(
                'CaptchaV2.verify: CAPTCHA_PRIVATE_KEY setting is unset.')

        if isinstance(captcha_solution, unicode):
            captcha_solution = captcha_solution.encode('utf-8')

        request_data = urllib.urlencode({
            'remoteip': remote_ip,
            'response': captcha_solution,
            'secret': settings.CAPTCHA_PRIVATE_KEY,
        })
        request = urllib2.Request(settings.CAPTCHA_V2_VERIFY_URL, request_data)
        with maybe_contextmanager(timer, 'captcha-verify',
                                  settings.CAPTCHA_V2_VERIFY_URL):
            self.response = self._open(request)

        if not self.response.is_error:
            try:
                response_data = json.loads(self.response.data())
                self._verified = response_data.get('success', False)
                self.message = 'ok'
            except ValueError:
                self._verified = False
        else:
            self._verified = False
            raise VerifyCaptchaError(self.response)
        return self._verified
