# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from contextlib import contextmanager
from cStringIO import StringIO

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from mock import Mock

from identityprovider.models.captcha import (
    CaptchaResponse,
    CaptchaV2,
    VerifyCaptchaError,
)
from identityprovider.tests.utils import SSOBaseTestCase


class CaptchaResponseTestCase(TestCase):

    def test_no_data(self):
        r = CaptchaResponse(200, None)
        self.assertEqual(r.data(), None)

    def test_data_from_response(self):
        resp = StringIO('this is the response')
        r = CaptchaResponse(200, resp)
        self.assertEqual(r._data, None)
        self.assertEqual(r.data(), 'this is the response')

    def test_data_exists(self):
        r = CaptchaResponse(200, None)
        r._data = 'this is the data'
        self.assertEqual(r.data(), 'this is the data')


@override_settings(CAPTCHA_PUBLIC_KEY='public', CAPTCHA_PRIVATE_KEY='private')
class CaptchaBaseTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CaptchaBaseTestCase, self).setUp()
        # patch Captcha so it never hits the real network
        self.captcha_response = CaptchaResponse(
            code=666, response=StringIO('\n'))
        self.mock_captcha_open = self.patch(
            'identityprovider.models.captcha.CaptchaBase._open',
            return_value=self.captcha_response)


class CaptchaV2TestCase(CaptchaBaseTestCase):

    def test_opener_is_created_during_init(self):
        CaptchaV2.opener = None
        CaptchaV2()
        self.assertTrue(CaptchaV2.opener is not None)

    def test_new_captcha_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield

        self.captcha_response.code = 200

        CaptchaV2(timer=timer_fn)

        self.assertEqual(1, len(passed_info))
        self.assertEqual('captcha-new', passed_info[0][0])

    def test_verify_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield
        c = CaptchaV2()
        c.verify(None, 'localhost', '', timer=timer_fn)
        self.assertEqual('captcha-verify', passed_info[0][0])
        self.assertEqual(settings.CAPTCHA_V2_VERIFY_URL, passed_info[0][1])


class CaptchaV2VerifyTestCase(CaptchaBaseTestCase):

    def test_verify_calls_check_whitelist(self):
        captcha = CaptchaV2(None)
        captcha.check_whitelist = Mock(return_value=False)

        result = captcha.verify(None, 'localhost', 'foo')
        self.assertFalse(result)
        self.assertTrue(self.mock_captcha_open.called)

        captcha.check_whitelist.assert_called_with('foo')

    def test_check_whitelist_match(self):
        captcha = CaptchaV2(None)

        regexps = ['not a match', '^foo$']
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=regexps):
            result = captcha.verify(None, 'localhost', 'foo')
            self.assertTrue(result)
            self.assertFalse(self.mock_captcha_open.called)

    def test_check_whitelist_no_match(self):
        captcha = CaptchaV2(None)

        regexps = ['not a match', '^bar$']
        with self.settings(EMAIL_WHITELIST_REGEXP_LIST=regexps):
            result = captcha.verify(None, 'localhost', 'foo')
            self.assertFalse(result)
            self.assertTrue(self.mock_captcha_open.called)

    def test_verify_already_verified(self):
        c = CaptchaV2(None)
        c._verified = True
        r = c.verify(None, None, '')
        self.assertTrue(r)

    def test_verify_response_ok(self):
        self.captcha_response.code = 200
        self.captcha_response.response = StringIO('{"success": true}')

        c = CaptchaV2()
        self.assertTrue(c.verify(None, "localhost", ''))
        self.assertEqual(c.message, 'ok')

    def test_verify_error(self):
        c = CaptchaV2(None)
        c.captcha_id = 'challenge-id'

        self.captcha_response.is_error = True
        self.assertRaises(VerifyCaptchaError, c.verify, None, 'localhost', '')

    def test_verify_uses_timer(self):
        passed_info = []

        @contextmanager
        def timer_fn(category, detail):
            passed_info.append((category, detail))
            yield
        c = CaptchaV2(None)
        c.verify(None, 'localhost', '', timer=timer_fn)
        self.assertEqual('captcha-verify', passed_info[0][0])
        self.assertEqual(
            settings.CAPTCHA_V2_VERIFY_URL, passed_info[0][1])
