# Copyright 2010-2019 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from base64 import b16encode
from time import time

from django.conf import settings
from oath import hotp, totp

from identityprovider.models import AuthenticationDevice
from identityprovider.tests.utils import SSOBaseTestCase


class DeviceTypeTestCase(SSOBaseTestCase):
    def setUp(self):
        super(DeviceTypeTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.key = b16encode('A' * 20)

    def create_device(self, device_type=None):
        return AuthenticationDevice.objects.create(
            account=self.account,
            name='Fake device',
            key=self.key,
            counter=0,
            device_type=device_type,
        )

    def verify_device(self, device_type, expected_actual_type,
                      expected_oath_type, expected_is_totp):
        device = self.create_device(device_type)
        self.assertEqual(device.device_type, device_type)
        self.assertEqual(device.actual_device_type, expected_actual_type)
        self.assertEqual(device.oath_type, expected_oath_type)
        self.assertEqual(device.is_totp(), expected_is_totp)

    def test_invalid_type(self):
        self.verify_device(device_type=None,
                           expected_actual_type=None,
                           expected_oath_type='hotp',
                           expected_is_totp=False)

    def test_empty_type(self):
        self.verify_device(device_type='',
                           expected_actual_type='',
                           expected_oath_type='hotp',
                           expected_is_totp=False)

    def test_generic_type(self):
        self.verify_device(device_type='generic',
                           expected_actual_type='generic',
                           expected_oath_type='hotp',
                           expected_is_totp=False)

    def test_empty_type_hotp(self):
        self.verify_device(device_type=':hotp',
                           expected_actual_type='',
                           expected_oath_type='hotp',
                           expected_is_totp=False)

    def test_empty_type_totp(self):
        self.verify_device(device_type=':totp',
                           expected_actual_type='',
                           expected_oath_type='totp',
                           expected_is_totp=True)

    def test_generic_type_hotp(self):
        self.verify_device(device_type='generic:hotp',
                           expected_actual_type='generic',
                           expected_oath_type='hotp',
                           expected_is_totp=False)

    def test_generic_type_totp(self):
        self.verify_device(device_type='generic:totp',
                           expected_actual_type='generic',
                           expected_oath_type='totp',
                           expected_is_totp=True)


class HOTPDeviceTestCase(SSOBaseTestCase):

    def setUp(self):
        super(HOTPDeviceTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.key = b16encode('A' * 20)
        self.device = self.factory.make_device(
            account=self.account, key=self.key, oath_mode='hotp')

    def test_authenticate_valid_dec6(self):
        otp = hotp(self.key, 0, 'dec6')
        self.assertTrue(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 1)

    def test_authenticate_valid_dec8(self):
        otp = hotp(self.key, 0, 'dec8')
        self.assertTrue(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 1)

    def test_authenticate_valid_with_max_drift(self):
        otp = hotp(self.key, settings.HOTP_DRIFT)
        self.assertTrue(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, settings.HOTP_DRIFT + 1)

    def test_authenticate_fails_with_max_plus_1_drift(self):
        otp = hotp(self.key, settings.HOTP_DRIFT + 1)
        self.assertFalse(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 0)

    def test_ordering(self):
        d1 = self.factory.make_device(self.account, name='foo')
        d2 = self.factory.make_device(self.account, name='bar')
        self.assertEqual(
            [x.id for x in AuthenticationDevice.objects.all().order_by('-id')],
            [x.id for x in d2, d1, self.device])
        self.assertEqual(
            [x.id for x in AuthenticationDevice.objects.all()],
            [x.id for x in self.device, d1, d2])

    def test_lp_1207398(self):
        otp = hotp('435126CCA7816847447EE16D0C49C30C219C0F62', 0, 'dec6')
        self.assertEqual(otp, '034801')


class TOTPDeviceTestCase(SSOBaseTestCase):

    def setUp(self):
        super(TOTPDeviceTestCase, self).setUp()
        self.account = self.factory.make_account()
        self.key = b16encode('A' * 20)
        self.device = self.factory.make_device(
            account=self.account, key=self.key, oath_mode='totp')

    def test_authenticate_valid_dec6(self):
        otp = totp(self.key, 'dec6')
        self.assertTrue(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 0)

    def test_authenticate_valid_dec8(self):
        otp = totp(self.key, 'dec8')
        self.assertTrue(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 0)

    def test_authenticate_invalid_dec6(self):
        otp = totp(self.key, 'dec6')
        self.assertFalse(self.device.authenticate(otp[1:]+otp[:1]))

    def test_authenticate_invalid_dec8(self):
        otp = totp(self.key, 'dec8')
        self.assertFalse(self.device.authenticate(otp[1:]+otp[:1]))

    def test_authenticate_with_just_expired_code_ok(self):
        # Calculate otp code with the previously valid, now expired code. Note
        # this test verifies we are accepting slightly old codes, which
        # requires TOTP_BACKWARD_DRIFT to be 1 or higher.
        # Baseline, aligned to 30-second (TOTP_PERIOD, really) boundary
        now = time()
        base = now - now % settings.TOTP_PERIOD
        # 1 second into the previous window
        the_past = base - 1
        # Generate this past, valid code and auth with it
        otp = totp(self.key, 'dec6', t=the_past)
        self.assertTrue(
            self.device.authenticate(otp),
            "This test WILL fail if TOTP_BACKWARD_DRIFT is 0. "
            "Double-check the setting and remove/skip this test "
            "if the desired behavior/setting has changed.")
        # A drifty client should NOT cause the device counter to
        # be updated (LP #1817075)
        self.assertEqual(self.device.counter, 0)

    def test_authenticate_with_slightly_too_new_code_ok(self):
        # Calculate otp code with the immediate next code. Note this test
        # verifies we are accepting slightly new codes, which
        # requires TOTP_FORWARD_DRIFT to be 1 or higher.
        # Baseline, aligned to 30-second (TOTP_PERIOD, really) boundary
        now = time()
        base = now - now % settings.TOTP_PERIOD
        # 1 second into the next window
        the_past = base + settings.TOTP_PERIOD + 1
        # Generate this future, valid code and auth with it
        otp = totp(self.key, 'dec6', t=the_past)
        self.assertTrue(
            self.device.authenticate(otp),
            "This test WILL fail if TOTP_FORWARD_DRIFT is 0. "
            "Double-check the setting and remove/skip this test "
            "if the desired behavior/setting has changed.")
        # A drifty client should NOT cause the device counter to
        # be updated (LP #1817075)
        self.assertEqual(self.device.counter, 0)

    def test_authenticate_with_too_old_code_fails(self):
        # Baseline, aligned to 30-second (TOTP_PERIOD, really) boundary
        now = time()
        base = now - now % settings.TOTP_PERIOD
        # Go TOTP_BACKWARD_DRIFT * TOTP_PERIOD seconds into the past,
        # which should still be valid...
        delta = settings.TOTP_PERIOD * settings.TOTP_BACKWARD_DRIFT
        the_past = base - delta
        # and now go 1 second into the past, which is now invalid
        the_past -= 1
        # Generate this past, invalid code and auth with it
        otp = totp(self.key, 'dec6', t=the_past)
        self.assertFalse(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 0)

    def test_authenticate_with_too_future_code_fails(self):
        # Baseline, aligned to 30-second (TOTP_PERIOD, really) boundary
        now = time()
        base = now - now % settings.TOTP_PERIOD
        # go TOTP_FORWARD_DRIFT * (TOTP_PERIOD + 1) seconds into the future,
        # which is invalid (the ending value is non-inclusive)
        delta = settings.TOTP_PERIOD * (settings.TOTP_FORWARD_DRIFT + 1)
        the_future = base + delta
        # Generate this future, invalid code and auth with it
        otp = totp(self.key, 'dec6', t=the_future)
        self.assertFalse(self.device.authenticate(otp))
        self.assertEqual(self.device.counter, 0)
