# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

import re
from base64 import b16encode
from collections import namedtuple
from urlparse import urlparse, urlunparse

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.views import redirect_to_login
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import ugettext_lazy as _
from django.views.generic import View
from gargoyle import gargoyle
from M2Crypto.Rand import rand_bytes
from oath import accept_hotp, accept_totp, hotp
from webauthn import (
        WebAuthnMakeCredentialOptions,
        WebAuthnRegistrationResponse,
)
from webauthn.webauthn import RegistrationRejectedException

from identityprovider.forms import DeviceRenameForm, OATHDeviceForm
from identityprovider.models import (
        AuthenticationDevice,
        U2FAuthenticationDevice,
)
from identityprovider.models import twofactor
from identityprovider.models.twofactor import get_otp_type

from webui.decorators import require_twofactor_enabled, sso_login_required
from webui.views.const import (
    DEVICE_ADDED,
    DEVICE_DELETED,
    DEVICE_GENERATION_WARNING,
    DEVICE_RENAMED,
    OTP_MATCH_ERROR,
)
from webui.views.utils import HttpResponseSeeOther, allow_only


DEVICE_ADDITION = 'device-addition'
DEVICE_LIST = 'device-list'
CodePageDetails = namedtuple('CodePageDetails', 'codes page start position')


def context_with_current_section(request, **kwargs):
    kwargs['current_section'] = 'devices'
    return kwargs


device_types = {
    'yubi': _('Yubikey'),
    'google': _('Google Authenticator'),
    'generic': _('Authentication device'),
    'paper': _('Printable Backup Codes'),
}

oath_modes = {
    'totp': _('Time Based'),
    'hotp': _('Counter Based'),
}


def generate_key(n):
    """Returns an OATH HOTP/TOTP key as a string of n raw bytes."""

    # An OATH HOTP/TOTP key is just bunch of random (in the "unpredictable"
    # sense) bits, of certain quantities (e.g. 160 bits or 20 bytes)
    # that are compatible with the AES algorithms.

    # From openssl's documentation:
    #
    #   RAND_bytes() puts num cryptographically strong pseudo-random
    #   bytes into buf. An error occurs if the PRNG has not been
    #   seeded with enough randomness to ensure an unpredictable byte
    #   sequence.
    #
    # openssl's RAND_bytes(num) function is available in Python as
    # M2Crypto.Rand.rand_bytes(num).

    return b16encode(rand_bytes(n))


def get_unique_device_name_for_user(name, user):
    return get_unique_device_name_for_user_for_qs(
        name,
        user,
        user.devices.all()
    )


def get_unique_u2f_device_name_for_user(name, user):
    return get_unique_device_name_for_user_for_qs(
        name,
        user,
        user.u2f_devices.all()
    )


def get_unique_device_name_for_user_for_qs(name, user, qs):
    """Returns the name with an extra number to make it unique if it exists in
       existing_names for that queryset. Objects in the queryset
       are assumed to have a name attribute.
    """
    original_name = name
    counter = 1
    existing_names = [device.name for device in qs]
    while name in existing_names:
        name = '%s (%d)' % (original_name, counter)
        counter += 1
    return name


@sso_login_required
@require_twofactor_enabled
@allow_only('GET')
def device_list(request, token=None):
    paper_renewals = list(request.user.paper_devices_needing_renewal)
    context = context_with_current_section(
        request, device_addition_path=reverse(DEVICE_ADDITION),
        devices=request.user.devices.all(),
        need_backup_device_warning=request.user.need_backup_device_warning,
        paper_devices_needing_renewal=paper_renewals,
        token=token
    )
    return render(request, 'device/list.html', context)


@sso_login_required
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_addition(request):
    if (request.user.has_twofactor_devices() and
            not twofactor.is_upgraded(request, must_be_fresh=True)):
        return redirect_to_login(request.get_full_path(), reverse('twofactor'))

    if request.method == 'GET':
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST))
        return render(request, 'device/types.html', context)

    device_type = request.POST.get('type')

    if device_type not in device_types.keys():
        return render(request, 'device/types.html',
                      context_with_current_section(request))

    if device_type == 'paper':
        return _device_addition_paper(request)
    return _device_addition_standard(request, device_type)


def _device_addition_paper(request):
    hex_key = generate_key(20)
    device_name = get_unique_device_name_for_user(device_types['paper'],
                                                  request.user)
    device = _create_device(request, device_name, hex_key, 0, 'paper')
    return HttpResponseSeeOther(reverse('device-print', args=(device.id,)))


def _device_addition_standard(request, device_type):
    error = None

    if 'hex_key' in request.POST:
        hex_key = request.POST.get('hex_key')
    else:
        # TODO: 20 bytes = 160 bits; this will change based on
        # device-type.
        hex_key = generate_key(20)

    if 'name' not in request.POST:
        initial_name = get_unique_device_name_for_user(
            device_types.get(device_type), request.user)
        form = OATHDeviceForm(initial={'name': initial_name})
    else:
        form = OATHDeviceForm(request.POST)
        if form.is_valid():
            device_name = get_unique_device_name_for_user(
                form.cleaned_data['name'], request.user)
            otp = form.cleaned_data['otp']
            otp_type = get_otp_type(otp)

            oath_type = 'totp'
            accepted, new_device_value = accept_totp(
                hex_key, str(otp), otp_type, period=settings.TOTP_PERIOD,
                forward_drift=settings.TOTP_FORWARD_DRIFT,
                backward_drift=settings.TOTP_BACKWARD_DRIFT, drift=0)

            if not accepted:
                oath_type = 'hotp'
                accepted, new_device_value = accept_hotp(
                    hex_key, str(otp), 0, otp_type, drift=settings.HOTP_DRIFT,
                    backward_drift=settings.HOTP_BACKWARDS_DRIFT)

            if accepted:
                _create_device(request, device_name, hex_key, new_device_value,
                               device_type, oath_type)
                return HttpResponseSeeOther(reverse(DEVICE_LIST))

            # Otherwise, set the error flag and fall through...
            error = OTP_MATCH_ERROR

    # Google would base32-encode, yubi would hex-encode, etc.  There
    # might even be multiple formats displayed simultaneously.
    formatted_key = re.sub('(.{4})', r'\1 ', hex_key).strip()

    context = context_with_current_section(
        request,
        device_list_path=reverse(DEVICE_LIST),
        type=device_type,
        oath_modes=oath_modes,
        ident="/".join([settings.TWOFACTOR_SERVICE_IDENT,
                        request.user.preferredemail.email]),
        hex_key=hex_key,
        form=form,
        formatted_key=formatted_key,
        error=error,
    )
    return render(request, 'device/addition-%s.html' % device_type, context)


def _create_device(
        request, device_name, hex_key, counter, device_type, oath_type=None):
    device = AuthenticationDevice.objects.create(
        account=request.user,
        name=device_name,
        key=hex_key,
        counter=counter,
        device_type=device_type + (
            ':{}'.format(oath_type) if oath_type else ''),
    )
    twofactor.login(request)
    added_type = oath_type.upper() if oath_type else device_types[device_type]
    messages.success(
        request, DEVICE_ADDED.format(
            name=device_name, type=added_type), 'temporary')
    return device


def _create_webauthn_device(
        request, device_name, ukey, public_key, credential_id, sign_count=0):
    device = U2FAuthenticationDevice.objects.create(
        account=request.user,
        name=device_name,
        ukey=ukey,
        pub_key=public_key,
        credential_id=credential_id,
        sign_count=sign_count,
    )
    twofactor.login(request)
    messages.success(
        request, DEVICE_ADDED.format(
            name=device_name, type="U2F"), 'temporary')
    return device


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET')
def device_print(request, device_id, token=None):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user,
        device_type='paper')

    details = _codes_for_position(device)
    remaining_codes = settings.TWOFACTOR_PAPER_CODES - details.position
    generation_enabled = (
        remaining_codes <= settings.TWOFACTOR_PAPER_CODES_ALLOW_GENERATION)
    needs_renewal = (
        remaining_codes < settings.TWOFACTOR_PAPER_CODES_WARN_RENEWAL)

    if generation_enabled:
        messages.warning(request, DEVICE_GENERATION_WARNING)

    context = context_with_current_section(
        request,
        codes=details.codes,
        counter=details.position,
        device_id=device.id,
        generation_enabled=generation_enabled,
        needs_renewal=needs_renewal,
    )
    if token is not None:
        context['token'] = token
    return TemplateResponse(request, 'device/print-codes.html', context)


def _codes_for_position(device, next_page=False):
    # use integer division to round the "window" boundaries
    page_size = settings.TWOFACTOR_PAPER_CODES
    page, page_position = divmod(device.counter, page_size)
    if next_page:
        page += 1
    page_start = page * page_size
    codes = [hotp(device.key, i, 'dec6')
             for i in range(page_start, page_start + page_size)]
    return CodePageDetails(codes, page, page_start, page_position)


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_generate(request, device_id, token=None):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user,
        device_type='paper')

    # find the next page of codes
    details = _codes_for_position(device, next_page=True)

    if request.method == 'GET':
        context = context_with_current_section(
            request,
            codes=details.codes,
            device_id=device.id,
            token=token,
        )
        return TemplateResponse(request, 'device/generate-codes.html', context)

    device.counter = details.start
    device.save()
    kwargs = {'device_id': device.id}
    if token:
        kwargs['token'] = token
    return HttpResponseRedirect(reverse('device-print', kwargs=kwargs))


@sso_login_required(require_twofactor=True, require_twofactor_freshness=True)
@require_twofactor_enabled
@allow_only('GET', 'POST')
def device_removal(request, device_id):
    device = get_object_or_404(
        AuthenticationDevice, id=device_id, account=request.user)

    if request.method != 'POST':
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), name=device.name)
        return render(request, 'device/removal.html', context)

    device.delete()

    # We should probably send an e-mail to the user stating which
    # device was removed. As a security measure, this would be much
    # stronger if bugs #784813, #784817, and #784818 were done.

    if not request.user.has_twofactor_devices():
        request.user.twofactor_required = False
        request.user.save()
        twofactor.logout(request)

    messages.success(request, DEVICE_DELETED.format(name=device.name))
    return HttpResponseSeeOther('/device-list')


class DeviceRenameView(View):
    def get(self, request, device_id):
        device = get_object_or_404(
            AuthenticationDevice, id=device_id, account=request.user)
        form = DeviceRenameForm({'name': device.name})
        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), form=form)
        return render(request, 'device/rename.html', context)

    def post(self, request, device_id):
        device = get_object_or_404(
            AuthenticationDevice, id=device_id, account=request.user)
        form = DeviceRenameForm(request.POST)
        if form.is_valid():
            original_name = device.name
            device.name = form.cleaned_data['name']
            device.save()
            messages.success(request,
                             DEVICE_RENAMED.format(original=original_name,
                                                   renamed=device.name))
            return HttpResponseRedirect(reverse(DEVICE_LIST))

        context = context_with_current_section(
            request, device_list_path=reverse(DEVICE_LIST), form=form)
        return render(request, 'device/rename.html', context)


device_rename = sso_login_required(
    require_twofactor=True,
    require_twofactor_freshness=True)(DeviceRenameView.as_view())


@allow_only('GET')
def device_help(request):
    if gargoyle.is_active('CAN_VIEW_SUPPORT_PHONE', request.user):
        support_phone = settings.SUPPORT_PHONE
    else:
        support_phone = ''

    context = {'support_phone': support_phone}
    return render(request, 'device/device-help.html', context)


# This view does not use sso_login_required (which does login and
# 2fa checks) because that decorator redirects the user to either
# 2fa verification or login, and in the API context we don't want that; we
# just want to check for those conditions and fail if not met. A happy-path
# request should come from a user that is already logged in and has fresh 2FA
# verification (if applicable).

class WebAuthnRegisterView(View):

    def post(self, request):
        if not gargoyle.is_active('WEBAUTHN', request):
            return JsonResponse(
                    {"error": "feature not enabled"},
                    status=404,
            )
        # Are we logged in? Only informative, the API consumer could redirect
        # the user to login but a non-logged-in user should really not be able
        # to hit this endpoint anyway, so best to just inform.
        if not request.user.is_authenticated():
            return JsonResponse({"error": "not logged in"}, status=400)
        # Check for existing 2-factor and fresh session (see comment below)
        # but DON'T redirect to 2fa check, because this is an API call which
        # the user won't see directly :(
        # The API consumer can opt to redirect the user if needed.
        if (request.user.has_twofactor_devices() and
                not twofactor.is_upgraded(request, must_be_fresh=True)):
            return JsonResponse(
                {"error": "current 2FA auth required"}, status=400)

        if request.POST == {}:
            # Empty post, generate and send challenge
            challenge = get_random_string(32)
            http_host = urlparse(
                request.build_absolute_uri()).netloc.split(":")[0]

            credential_options = WebAuthnMakeCredentialOptions(
                challenge=challenge,
                rp_name="Ubuntu One SSO",
                # VERY IMPORTANT: the RP_ID must MATCH the URL/HOSTNAME.
                # Otherwise the browser will spit out Error creating
                # credential: DOMException: "The operation is insecure." If you
                # get this error and expand it, it'll tell you which host it
                # was expecting.
                rp_id=http_host,
                user_id=request.user.openid_identifier,
                username=request.user.username,
                display_name=request.user.displayname,
                icon_url="",
                attestation='indirect',
            )
            # Write this in the session so we can retrieve and check it later
            request.session['challenge'] = challenge
            request.session['ukey'] = request.user.openid_identifier

            return JsonResponse(credential_options.registration_dict)
        else:
            # Non-empty post, assume it's a response, verify and maybe register
            # device.
            if not ('challenge' in request.session
                    or 'ukey' in request.session):
                return JsonResponse(
                    {"error": "incomplete registration session data"},
                    status=400)
            if not set(
                    ("id", "type", "rawId", "attObj", "clientData",
                     "registrationClientExtensions",)).issubset(
                        request.POST.keys()):
                return JsonResponse(
                    {"error": "incomplete registration form data"},
                    status=400)
            # Must massage the form into registration_response
            registration_response = request.POST
            ukey = request.session['ukey']
            challenge = request.session['challenge']
            parsed_url = urlparse(request.build_absolute_uri())
            # Http_host is just the hostname (e.g. login.ubuntu.com)
            http_host = parsed_url.netloc.split(":")[0]
            # Origin is the base URL (e.g. https://login.ubuntu.com)
            origin = urlunparse(parsed_url[:2] + ("", "", "", ""))

            webauthn_registration_response = WebAuthnRegistrationResponse(
                rp_id=http_host,
                origin=origin,
                registration_response=registration_response,
                challenge=challenge,
                trust_anchor_dir="u2f-trust-roots",
                trusted_attestation_cert_required=True,
                self_attestation_permitted=True,
                none_attestation_permitted=True,  # Can be False w/direct att.
                uv_required=False)  # User Verification
            try:
                webauthn_credential = webauthn_registration_response.verify()
            except RegistrationRejectedException as e:
                # verify() response is an exception whose repr looks like:
                # Registration rejected. Error: FOOOO.
                # (where FOOOO is the repr of the *actual* exception.)
                return JsonResponse({'error': '{}'.format(e)}, status=400)

            # Default name if none given in the form
            if 'device_name' not in request.POST:
                device_name = get_unique_device_name_for_user(
                    'U2F', request.user)
            else:
                device_name = request.POST['device_name']
            # Create the device; the method takes care of setting
            # a flash message for the user.
            _create_webauthn_device(
                request, device_name=device_name, ukey=ukey,
                public_key=webauthn_credential.public_key,
                credential_id=webauthn_credential.credential_id,
                sign_count=webauthn_credential.sign_count)
            # Client app should notice this response and redirect victim
            return JsonResponse({"yay": "reg validation success"})


webauthn_register = WebAuthnRegisterView.as_view()
